#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2019 Cyril Desjouy <cyril.desjouy@univ-lemans.fr>
#
# This file is part of nsfds2
#
# nsfds2 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nsfds2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nsfds2. If not, see <http://www.gnu.org/licenses/>.
#
# Creation Date : 2019-05-02 - 14:09:56
#
# pylint: disable=too-many-locals
# pylint: disable=too-many-statements
# pylint: disable=too-many-instance-attributes

"""
-----------
DOCSTRING

@author: Cyril Desjouy
"""

import os
import sys
import getpass
import h5py
import numpy as _np
import matplotlib.pyplot as _plt
import matplotlib.animation as _ani
from mpl_toolkits.axes_grid1 import make_axes_locatable
from ofdlib2 import fdtd as _fdtd
from progressbar import ProgressBar, Bar, ReverseBar, ETA
from mplutils import modified_jet, MidpointNormalize
from fdgrid.mesh import plot_subdomains as _plt_subdomains


__all__ = ['get_data', 'show', 'fields', 'probes', 'FrameGenerator', 'Movie']


def get_data(filename):
    """ Get data from filename. """

    try:
        data = h5py.File(filename, 'r')
    except OSError:
        print('You must provide a valid hdf5 file : moviemaker mydir/myfile.hdf5')
        sys.exit(1)
    else:
        return data


def show():
    """ Show all figures. """
    _plt.show()


def fields(cfg):
    """ Make figure """

    if not cfg.figures or not cfg.save:
        return None

    with h5py.File(cfg.datafile, 'r') as sfile:

        nt = sfile['nt'][...]
        x = sfile['x'][...]
        z = sfile['z'][...]
        obstacles = sfile['obstacles'][...]
        if cfg.mesh == 'curvilinear':
            J = sfile['J'][...]
        else:
            J = _np.ones((x.size, z.size))

        if cfg.onlyp:
            p = sfile[f'p_it{nt}'][...]
        else:
            rho = sfile[f'rho_it{nt}'][...]*J
            rhou = sfile[f'rhou_it{nt}'][...]*J
            rhov = sfile[f'rhov_it{nt}'][...]*J
            rhoe = sfile[f'rhoe_it{nt}'][...]*J
            u = rhou/rho
            v = rhov/rho
            e = rhoe/rho
            p = _np.empty_like(rho)
            _fdtd.p(p, rho, rhou, rhov, rhoe, cfg.gamma)


    cm = modified_jet()
    norm = MidpointNormalize(midpoint=0)

    if cfg.onlyp:
        _, ax = _plt.subplots(figsize=(12, 9))
        im = ax.pcolorfast(x, z, (p-cfg.p0).T, cmap=cm, norm=norm)
        _plt_subdomains(ax, x, z, obstacles)
        ax.set_xlabel(r'$x$ [m]')
        ax.set_ylabel(r'$z$ [m]')
        ax.set_aspect('equal')
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        _plt.colorbar(im, cax=cax)
        if cfg.probes:
            ax.plot(*cfg.probes_loc, 'ro')

    else:
        _, axes = _plt.subplots(2, 2, figsize=(12, 9))

        im1 = axes[0, 0].pcolorfast(x, z, (p-cfg.p0).T, cmap=cm, norm=norm)
        im2 = axes[0, 1].pcolorfast(x, z, u.T, cmap=cm)
        im3 = axes[1, 0].pcolorfast(x, z, v.T, cmap=cm)
        im4 = axes[1, 1].pcolorfast(x, z, e.T, cmap=cm)
        ims = [im1, im2, im3, im4]

        for ax, im in zip(axes.ravel(), ims):
            _plt_subdomains(ax, x, z, obstacles)
            ax.set_xlabel(r'$x$ [m]')
            ax.set_ylabel(r'$z$ [m]')
            ax.set_aspect('equal')
            divider = make_axes_locatable(ax)
            cax = divider.append_axes("right", size="5%", pad=0.05)
            _plt.colorbar(im, cax=cax)
            if cfg.probes and cfg.probes_loc:
                ax.plot(*cfg.probes_loc, 'ro')

    return None


def probes(cfg):
    """ Plot probes. """

    if cfg.figures and cfg.probes:

        with h5py.File(cfg.datafile, 'r') as sfile:

            nt = sfile['nt'][...]
            dt = sfile['dt'][...]
            p0 = sfile['p0'][...]
            pressure = sfile['probes'][...] - p0
            probes_loc = sfile['probes_location'][...]

        t = _np.arange(nt)*dt

        _, ax = _plt.subplots(figsize=(9, 4))
        for i, c in enumerate(probes_loc):
            ax.plot(t, pressure[i, :], label=f'@{tuple(c)}')
        ax.set_xlim(t.min(), t.max())
        ax.set_xlabel('Time [s]')
        ax.set_ylabel('Pressure [Pa]')
        ax.legend()
        ax.grid()


class FrameGenerator:
    """ Frame Genrator

    Parameters
    ----------

    data : Path to hdf5 file (str) or data from hdf5 file
    view : The variable to display. string.
    ref : The reference frame for colormap. int.
    nt : The last frame to consider. int

    """

    def __init__(self, data, view='p', ref=20, nt=None):

        if isinstance(data, str):
            self.data = get_data(data)
        else:
            self.data = data
        self.view = view
        self.ref = ref
        self.nt = nt
        self.ns = data['ns'][...]
        self.icur = 0
        self.nx = self.data['nx'][...]
        self.nz = self.data['nz'][...]
        if self.data['mesh'][...] == 'curvilinear':
            self.J = self.data['J'][...]
        else:
            self.J = _np.ones((self.nx, self.nz))

    def reference(self):
        """ Generate the reference for min/max colormap values """

        if self.view == "p":
            ref = self.compute_p(self.ref)
        elif self.view in ['rho', 'vx', 'vz', 'e']:
            ref = self.data[f"{self.view}_it{self.ref}"][...]
        else:
            print("Only 'p', 'rho', 'vx', 'vz' and 'e' available !")
            sys.exit(1)

        return (ref*self.J).max(), (ref*self.J).min()

    def compute_p(self, it):
        """ Compute p from rho, rhou, rhov and rhoe at iteration it. """

        rho = self.data[f"rho_it{it}"][...]*self.J
        rhou = self.data[f"rhou_it{it}"][...]*self.J
        rhov = self.data[f"rhov_it{it}"][...]*self.J
        rhoe = self.data[f"rhoe_it{it}"][...]*self.J
        p = _np.empty_like(rho)
        _fdtd.p(p, rho, rhou, rhov, rhoe, self.data['gamma'][...])

        return p - self.data['p0'][...]

    def _next_item(self):
        """ Generate next value of variable """

        if self.view == 'p':
            return self.compute_p(self.icur).T

        if self.view == 'rho':
            return self.data[f"{self.view}_it{self.icur}"][...].T

        if self.view in ['vx', 'vz', 'e']:
            vx = self.data[f"{self.view}_it{self.icur}"][...].T
            rho = self.data[f"rho_it{self.icur}"][...].T
            return vx/rho

        return None

    def __iter__(self):
        """ Iterator """

        return self

    def __next__(self):
        """ Next element of iterator : (frame_number, variable) """

        try:
            self.icur += self.ns
            if self.nt:
                if self.icur > self.nt:
                    raise StopIteration

            return self.icur, self._next_item()

        except KeyError:
            raise StopIteration


class Movie:
    """ Make movie from nsfds2 hdf5 ouput file.

    Parameters
    ----------

    filename : hdf5 file
    view : The variable to display. string.
    ref : The reference frame for colormap. int.
    nt : The last frame to consider. int.
    quiet : quiet mode. Boolean.
    """

    def __init__(self, filename, view='p', ref=None, nt=None, quiet=False):

        self.filename = filename
        self.path = os.path.dirname(filename) + '/'
        self.data = get_data(filename)
        self.view = view
        self.ref = ref
        self.nt = nt
        self.quiet = quiet
        self.curvilinear = True if self.data['mesh'][...] == 'curvilinear' else False
        self.obstacles = self.data['obstacles'][...]

        self.check()
        self._init_coordinates()
        self._init_generator()
        self._init_cmap()
        self._init_pbar()
        self._init_movie()

    def check(self):
        """ Check parameters. """

        if not self.nt:
            self.nt = self.data['nt'][...]

        if not self.ref:
            self.ref = int(self.nt/2)

    def _init_coordinates(self):
        """ Init coordinate system. """

        if self.curvilinear:
            self.x = self.data['xp'][...]
            self.z = self.data['zp'][...]
        else:
            self.x = self.data['x'][...]
            self.z = self.data['z'][...]

    def _init_generator(self):
        """ Init FrameGenerator. """

        try:
            self.frames = FrameGenerator(self.data, self.view, ref=self.ref, nt=self.nt)
            self.maxp, self.minp = self.frames.reference()
        except KeyError:
            self.frames = FrameGenerator(self.data, self.view, nt=self.nt)
            self.maxp, self.minp = self.frames.reference()

    def _init_cmap(self):
        """ Init cmap. """

        self.mycm = modified_jet()
        if self.minp < 0:
            self.norm = MidpointNormalize(vmax=self.maxp, vmin=self.minp, midpoint=0)
        else:
            self.norm = None

    def _init_pbar(self):
        """ Init progress bar. """

        if not self.quiet:
            widgets = [Bar('>'), ' ', ETA(), ' ', ReverseBar('<')]
            self.pbar = ProgressBar(widgets=widgets, maxval=self.nt).start()

    def _init_movie(self):
        """ Init movie parameters. """

        title = os.path.basename(self.filename).split('.')[0]
        metadata = dict(title=title, artist=getpass.getuser(), comment='From nsfds2')
        self.writer = _ani.FFMpegWriter(fps=24, metadata=metadata, bitrate=-1, codec="libx264")
        self.movie_filename = '{}.mkv'.format(title)

    def make(self):
        """ Make movie. """

        # 1st frame
        _, p = next(self.frames)
        movie = _plt.figure(figsize=(20, 9))
        movie.subplots_adjust(left=0.1, right=0.9, top=0.9, bottom=0.15)
        axm = movie.add_subplot(111)
        # Labels
        title = r'{} -- iteration : {}'
        axm.set_xlabel(r'$x$ [m]', fontsize=22)
        axm.set_ylabel(r'$y$ [m]', fontsize=22)
        axm.set_title(title.format(0, self.view))
        axm.set_aspect('equal')
        # plot
        if self.curvilinear:
            movie_plt = axm.pcolormesh(self.x, self.z, p.T, cmap=self.mycm, norm=self.norm)
            axm.plot(self.x[0, :], self.z[0, :], 'k', linewidth=3)
            axm.plot(self.x[-1, :], self.z[-1, :], 'k', linewidth=3)
            axm.plot(self.x[:, 0], self.z[:, 0], 'k', linewidth=3)
            axm.plot(self.x[:, -1], self.z[:, -1], 'k', linewidth=3)
            _plt_subdomains(axm, self.x, self.z, self.obstacles, edgecolor='k', curvilinear=True)
        else:
            movie_plt = axm.pcolorfast(self.x, self.z, p, cmap=self.mycm, norm=self.norm)
            _plt_subdomains(axm, self.x, self.z, self.obstacles)

        _plt.colorbar(movie_plt)

        # Start Video
        with self.writer.saving(movie, self.path + self.movie_filename, dpi=100):
            for i, var in self.frames:
                axm.set_title(r'{} -- iteration : {}'.format(self.view, i))
                if self.curvilinear:
                    # StackOv : using-set-array-with-pyplot-pcolormesh-ruins-figure
                    movie_plt.set_array(var[:-1, :-1].T.ravel())
                else:
                    movie_plt.set_data(var)
                self.writer.grab_frame()
                if not self.quiet:
                    self.pbar.update(i)
            if not self.quiet:
                self.pbar.finish()
        _plt.show()



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2019 Cyril Desjouy <cyril.desjouy@univ-lemans.fr>
#
# This file is part of nsfds2
#
# nsfds2 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nsfds2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nsfds2. If not, see <http://www.gnu.org/licenses/>.
#
#
# Creation Date : 2016-11-29 - 23:18:27
#
# pylint: disable=too-many-statements
"""
-----------

Parse config file and set all simulation parameters

@author: Cyril Desjouy
"""

import json
import time
import sys
import os
import configparser


class CfgSetup:
    """ Handle configuration file. """

    # pylint: disable=too-many-instance-attributes

    def __init__(self, args=None):

        # Command line arguments
        self.args = args

        # Create config parser
        self.cfg = configparser.RawConfigParser()
        self.home = os.path.expanduser("~")
        self.path = self.home + '/.nsfds2/'

        # Check if config dir exists. If not create it.
        self.check_dir(self.path)

        # Check if config file exist. If not create it
        self.init_cfg()

        # Read config file (can be overridden by command line)
        cf = getattr(self.args, 'cfgfile', None)
        if cf:
            self.cfg.read(cf)
        else:
            self.cfg.read(self.path + 'nsfds2.conf')

        # Parse arguments
        self.run()

    @staticmethod
    def check_dir(directory):
        """ Check if dir exists. If not, create it."""

        if not os.path.isdir(directory):
            os.makedirs(directory)
            print("Create directory :", directory)
            time.sleep(0.5)

    def init_cfg(self):
        """ Check if nsfds2.conf exists. If not create it. """

        if not os.path.exists(self.path + 'nsfds2.conf'):
            open(self.path + 'nsfds2.conf', 'a').close()
            print("Create configuration file : {}nsfds2.conf".format(self.path))
            time.sleep(0.5)
            self.write_default()

    def write_default(self):
        """ Write default configuration file. """

        self.cfg.add_section('configuration')
        self.cfg.set('configuration', 'timings', 'False')
        self.cfg.set('configuration', 'quiet', 'False')

        self.cfg.add_section('simulation')
        self.cfg.set('simulation', 'nt', '500')
        self.cfg.set('simulation', 'ns', '10')
        self.cfg.set('simulation', 'CFL', '0.5')

        self.cfg.add_section('geometry')
        self.cfg.set('geometry', 'mesh', 'regular')
        self.cfg.set('geometry', 'file', 'None')
        self.cfg.set('geometry', 'geoname', 'helmholtz_double')
        self.cfg.set('geometry', 'curvname', 'None')
        self.cfg.set('geometry', 'bc', 'PPPP')
        self.cfg.set('geometry', 'nx', '256')
        self.cfg.set('geometry', 'nz', '256')
        self.cfg.set('geometry', 'ix0', '0')
        self.cfg.set('geometry', 'iz0', '0')
        self.cfg.set('geometry', 'dx', '1')
        self.cfg.set('geometry', 'dz', '1')

        self.cfg.add_section('PML')
        self.cfg.set('PML', 'beta', 0.)
        self.cfg.set('PML', 'alpha', 4.)
        self.cfg.set('PML', 'sigmax', 20.)
        self.cfg.set('PML', 'sigmaz', 20.)
        self.cfg.set('PML', 'Npml', 15)

        self.cfg.add_section('source')
        self.cfg.set('source', 'type', 'pulse')
        self.cfg.set('source', 'ixS', '32')
        self.cfg.set('source', 'izS', '128')
        self.cfg.set('source', 'S0', '1e6')

        self.cfg.add_section('eulerian fluxes')
        self.cfg.set('eulerian fluxes', 'stencil', '11')

        self.cfg.add_section('filtering')
        self.cfg.set('filtering', 'filter', 'True')
        self.cfg.set('filtering', 'stencil', '11')
        self.cfg.set('filtering', 'stength', '0.75')

        self.cfg.add_section('viscous fluxes')
        self.cfg.set('viscous fluxes', 'viscosity', 'True')
        self.cfg.set('viscous fluxes', 'stencil', '7')

        self.cfg.add_section('shock capture')
        self.cfg.set('shock capture', 'shock capture', 'True')
        self.cfg.set('shock capture', 'stencil', '7')
        self.cfg.set('shock capture', 'method', 'pressure')

        self.cfg.add_section('figures')
        self.cfg.set('figures', 'figures', 'True')

        self.cfg.add_section('save')
        self.cfg.set('save', 'save', 'True')
        self.cfg.set('save', 'path', 'results/')
        self.cfg.set('save', 'filename', 'tmp')
        self.cfg.set('save', 'compression', 'lzf')
        self.cfg.set('save', 'only p', 'False')
        self.cfg.set('save', 'probes', 'False')
        self.cfg.set('save', 'probes_locations', '[]')

        with open(self.path + 'nsfds2.conf', 'w') as cf:
            self.cfg.write(cf)

    def run(self):
        """ Run configuration. """

        # Thermophysic parameters
        self.gamma = 1.4
        self.nu = 1.5e-5
        self.c0 = 340.
        self.rho0 = 1.22
        self.p0 = self.rho0*self.c0**2/self.gamma

        try:
            CFG = self.cfg['configuration']

            self.timings = getattr(self.args, 'timings', None)
            if not isinstance(self.timings, bool):
                self.timings = CFG.getboolean('timings', False)

            self.quiet = getattr(self.args, 'quiet', None)
            if not isinstance(self.quiet, bool):
                self.quiet = CFG.getboolean('quiet', False)

            SIM = self.cfg['simulation']
            self.nt = getattr(self.args, 'nt', None)
            if self.nt is None:
                self.nt = SIM.getint('nt', 500)
            self.ns = SIM.getint('ns', 10)
            self.CFL = SIM.getfloat('CFL', 0.5)

            GEO = self.cfg['geometry']
            self.mesh = GEO.get('mesh', 'regular')
            self.curvflag = True if self.mesh == 'curvilinear' else False
            self.geofile = getattr(self.args, 'geofile', None)
            self.geoflag = True
            if  self.geofile:
                self.geofile, self.geoname = self.args.geofile
            else:
                self.geofile = GEO.get('file', 'None')
                self.geoname = GEO.get('geoname', 'square')
            self.curvname = GEO.get('curvname', 'None')
            self.bc = GEO.get('bc', 'RRRR')
            self.nx = GEO.getint('nx', 256)
            self.nz = GEO.getint('nz', 256)
            self.ix0 = GEO.getint('ix0', 0)
            self.iz0 = GEO.getint('iz0', 0)
            self.dx = GEO.getfloat('dx', 1)
            self.dz = GEO.getfloat('dz', 1)
            self.dt = min(self.dx, self.dz)*self.CFL/self.c0

            PML = self.cfg['PML']
            self.beta = PML.getfloat('beta', 0.)
            self.alpha = PML.getfloat('alpha', 4.)
            self.sigmax = PML.getfloat('sigmax', 20.)
            self.sigmaz = PML.getfloat('sigmaz', 20.)
            self.Npml = PML.getint('Npml', 15)

            SRC = self.cfg['source']
            self.typ = SRC.get('type', 'pulse')
            self.ixS = SRC.getint('ixS', 32)
            self.izS = SRC.getint('izS', 32)
            self.S0 = SRC.getfloat('S0', 1e3)

            EUL = self.cfg['eulerian fluxes']
            self.stencil = EUL.getint('stencil', 11)

            FLT = self.cfg['filtering']
            self.flt = FLT.getboolean('filter', True)
            self.flt_stencil = FLT.getint('stencil', 11)
            self.xnu = FLT.getfloat('strength', 0.75)

            VSC = self.cfg['viscous fluxes']
            self.vsc = VSC.getboolean('viscosity', True)
            self.vsc_stencil = VSC.getint('stencil', 3)

            CPT = self.cfg['shock capture']
            self.cpt = CPT.getboolean('shock capture', True)
            self.cpt_stencil = CPT.getint('stencil', 7)
            self.cpt_meth = CPT.get('method', 'pressure')
            self.rth = 1e-6

            SAVE = self.cfg['save']
            self.save = SAVE.getboolean('save', True)
            self.savepath = SAVE.get('path', 'results/')
            self.savefile = SAVE.get('filename', 'tmp') + '.hdf5'
            self.comp = SAVE.get('compression', 'lzf')
            self.onlyp = SAVE.getboolean('only p', False)
            self.probes = SAVE.getboolean('probes', False)
            self.probes_loc = json.loads(SAVE.get('probes_locations', '[]'))

            # datapath and datafile
            if self.comp == 'None':
                self.comp = None

            if self.savepath and not self.savepath.endswith('/'):
                self.savepath += '/'

            # if self.savepath does not exist, create it
            self.check_dir(self.savepath)

            self.datafile = getattr(self.args, 'datafile', None)
            if not self.datafile:
                self.datafile = self.savepath + self.savefile

            FIGS = self.cfg['figures']
            self.figures = FIGS.getboolean('figures', True)

        except configparser.Error as err:
            print('Bad cfg file : ', err)
            sys.exit(1)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2019 Cyril Desjouy <cyril.desjouy@univ-lemans.fr>
#
# This file is part of nsfds2
#
# nsfds2 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nsfds2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nsfds2. If not, see <http://www.gnu.org/licenses/>.
#
# Creation Date : 2019-04-18 - 15:23:05
"""
-----------
DOCSTRING

@author: Cyril Desjouy
"""

import re
import itertools

def bc_combinations():
    """ List of possible combinations for bc"""

    corners = list(itertools.product(['X', 'A'], repeat=4))
    corners = [''.join(i) for i in corners if i.count('A') == 2]
    corners = [i for i in corners if any(re.match(regex, i) for regex in [r'A..A', r'AA..', r'..AA', r'.AA.'])]

    pml = list(itertools.product(['R', 'P', 'X', 'A'], repeat=4))
    pml = [''.join(i) for i in pml if i.count('A') == 1]
    pml = [i for i in pml if 'A' in i and i.count('P') < 2 and i.count('R') < 3 and i.count('X') > 0]
    pml = [i for i in pml if any(re.match(regex, i) for regex in [r'.A.X', r'.X.A', r'X.A.', r'A.X.'])]


    pml_p = [i for i in pml if 'P' in i]
    pml_r = [i for i in pml if 'R' in i and 'P' not in i]
    pml_x = [i for i in pml if 'R' not in i and 'P' not in i]

    print('\n* {} Corners : {}'.format(len(corners), ', '.join(corners)))
    print('\n* {} with only X : {}'.format(len(pml_x), ', '.join(pml_x)))
    print('\n* {} with P : {}'.format(len(pml_p), ', '.join(pml_p)))
    print('\n* {} with only R & X : {}'.format(len(pml_r), ', '.join(pml_r)))



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2019 Cyril Desjouy <cyril.desjouy@univ-lemans.fr>
#
# This file is part of nsfds2
#
# nsfds2 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nsfds2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nsfds2. If not, see <http://www.gnu.org/licenses/>.
#
#
# Creation Date : 2019-03-01 - 12:05:08
"""
-----------

Navier Stokes Finite Differences Solver

@author: Cyril Desjouy
"""

import os
import argparse
from fdgrid import mesh
from nsfds2.init import CfgSetup, Fields
from nsfds2.fdtd import FDTD
from nsfds2.utils import files, headers, graphics


def parse_args():
    """ Parse arguments. """

    parser = argparse.ArgumentParser(prog='nsfds',
                                     description='A Navier-Stokes Finite Difference Solver')

    subparsers = parser.add_subparsers(dest='command')
    solve_parser = subparsers.add_parser("solve")
    movie_parser = subparsers.add_parser("movie")
    show_parser = subparsers.add_parser("show")

    # Common
    for p in [solve_parser, show_parser]:
        p.add_argument('-g', '--geo-file', metavar='GF', dest='geofile', nargs=2,
                       help='file and pattern to use')
        p.add_argument('-c', '--cfg-file', metavar='CF', dest='cfgfile',
                       help='path to config file')

    for p in [solve_parser, movie_parser]:
        p.add_argument('-q', '--quiet', action="store_true", help='Quiet mode')
        p.add_argument('-i', dest='nt', type=int,
                       help='Number of time iterations')
        p.add_argument('-d', '--dat-file', metavar='DF', dest='datafile',
                       help='path to hdf5 data file')


    # Movie parser
    movie_parser.add_argument('view', nargs='?', default='p',
                              choices=['p', 'rho', 'vx', 'vz', 'e'])
    movie_parser.add_argument('-r', dest='ref', type=int,
                              help='Reference frame for colormap')

    # Show parser
    show_parser.add_argument('view', nargs='?', default='grid',
                             choices=['grid', 'domains', 'all'])

    # Solver parser
    solve_parser.add_argument('-t', '--timings', action="store_true", default=None,
                              help='Display complete timings')

    return parser.parse_args()


def solve(cfg, msh):
    """ Solve NS equations. """

    # Simulation parameters
    fld = Fields(msh, cfg)

    # Simulation
    fdtd = FDTD(msh, fld, cfg)
    fdtd.run()

    # Figures
    graphics.fields(cfg)
    graphics.probes(cfg)
    graphics.show()


def show(cfg, msh):
    """ Show simulation parameters and grid. """

    headers.check_geo(cfg)
    headers.parameters(cfg)

    if cfg.args.view == 'grid':
        msh.plot_grid()

    elif cfg.args.view == 'domains':
        print('\n')
        msh.plot_domains(legend=True)

    elif cfg.args.view == 'all':
        print('\n')
        msh.plot_grid()
        msh.plot_domains(legend=True)

    msh.show_figures()


def movie(cfg, _):
    """ Create a movie from a dataset. """

    movie = graphics.Movie(cfg.datafile, view=cfg.args.view,
                           ref=cfg.args.ref, nt=cfg.nt, quiet=cfg.quiet)
    movie.make()


def main():
    """ Main """

    # Parse arguments
    args = parse_args()

    # Parse config file
    cfg = CfgSetup(args=args)

    # Geometry
    obstacles = files.get_obstacle(cfg)

    # Mesh
    if cfg.mesh == 'regular':
        msh = mesh.Mesh((cfg.nx, cfg.nz), (cfg.dx, cfg.dz), origin=(cfg.ix0, cfg.iz0),
                        bc=cfg.bc, obstacles=obstacles, Npml=cfg.Npml,
                        stencil=cfg.stencil)
    elif cfg.mesh == 'curvilinear':
        fcurv = files.get_curvilinear(cfg)
        msh = mesh.CurvilinearMesh((cfg.nx, cfg.nz), (cfg.dx, cfg.dz),
                                   origin=(cfg.ix0, cfg.iz0),
                                   bc=cfg.bc, obstacles=obstacles, Npml=cfg.Npml,
                                   stencil=cfg.stencil, fcurvxz=fcurv)

    if args.command:
        globals()[args.command](cfg, msh)
    else:
        headers.copyright()
        print('Must specify an action among solve/movie/show')
        print('See nsfds2 -h for help')


if __name__ == "__main__":

    os.nice(20)
    main()



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2016-2019 Cyril Desjouy <cyril.desjouy@univ-lemans.fr>
#
# This file is part of nsfds2
#
# nsfds2 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nsfds2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nsfds2. If not, see <http://www.gnu.org/licenses/>.
#
#
# Creation Date : 2019-03-01 - 14:09:42
#
# pylint: disable=too-many-instance-attributes
"""
-----------

Initialize all fields

@author: Cyril Desjouy
"""

import h5py
import numpy as np
from ofdlib2.fdtd import fdtools

class Fields:
    """ Fields initialization. """

    def __init__(self, msh, cfg):

        self._msh = msh
        self._cfg = cfg
        self.fdtools = fdtools(self._cfg.nx, self._cfg.nz,
                               self._cfg.p0, self._cfg.gamma, self._cfg.dt)

        self.init_fields()
        self.init_derivatives()
        self.init_filters()
        if self._cfg.save:
            self.init_save()
        if 'A' in self._msh.bc:
            self.init_pml()

    def init_fields(self):
        """ Setup initial fields. """

        self.p = np.zeros(self._msh.shape)
        self.r = np.zeros_like(self.p) + self._cfg.rho0
        self.ru = np.zeros_like(self.p)
        self.rv = np.zeros_like(self.p)
        self.dltn = np.zeros_like(self.p)

        # Location
        ixS = self._cfg.ixS
        izS = self._cfg.izS
        Bx = 5*self._msh.dx                            # spatial bandwidth
        S0 = self._cfg.S0
        p0 = self._cfg.p0

        for iz in range(self._msh.nz):
            self.p[:, iz] = p0 + \
                    S0*np.exp(-np.log(2)*((self._msh.x-self._msh.x[ixS])**2 +
                                          (self._msh.z[iz]-self._msh.z[izS])**2)/Bx**2)


        if self._cfg.mesh == 'curvilinear':
            self.r = self.r/self._msh.J
            self.ru = self.ru/self._msh.J
            self.rv = self.rv/self._msh.J
            self.p = self.p/self._msh.J

        self.re = self.p/(self._cfg.gamma-1.)

    def init_derivatives(self):
        """ Init derivatives. """

        self.K = np.zeros_like(self.p)
        self.Ku = np.zeros_like(self.p)
        self.Kv = np.zeros_like(self.p)
        self.Ke = np.zeros_like(self.p)

        self.E = np.empty_like(self.p)
        self.Eu = np.empty_like(self.p)
        self.Ev = np.empty_like(self.p)
        self.Ee = np.empty_like(self.p)

        self.F = np.empty_like(self.p)
        self.Fu = np.empty_like(self.p)
        self.Fv = np.empty_like(self.p)
        self.Fe = np.empty_like(self.p)

    def init_filters(self):
        """ Init variables used for filtering. """

        self.dp = np.zeros_like(self.p)
        self.sg = np.zeros_like(self.p)
        self.tau11 = np.zeros_like(self.p)
        self.tau22 = np.zeros_like(self.p)
        self.tau12 = np.zeros_like(self.p)

    def init_pml(self):
        """ Init PMLs. """

        # sigmax
        Dx = self._msh.Npml*self._msh.dx                      # Width of the PML
        self.sx = np.zeros(self._msh.nx)
        self.sx[:self._msh.Npml] = self._cfg.sigmax*abs((self._msh.x[:self._msh.Npml] \
                                 - self._msh.x[self._msh.Npml])/Dx)**self._cfg.alpha
        self.sx[self._msh.nx - self._msh.Npml:] = self.sx[self._msh.Npml-1::-1]

        # sigmaz
        Dz = self._msh.Npml*self._msh.dz                      # Width of the PML
        self.sz = np.zeros(self._msh.nz)
        self.sz[:self._msh.Npml] = self._cfg.sigmaz*abs((self._msh.z[:self._msh.Npml] \
                                 - self._msh.z[self._msh.Npml])/Dz)**self._cfg.alpha
        self.sz[self._msh.nz - self._msh.Npml:] = self.sz[self._msh.Npml-1::-1]

        # Init Q
        self.qx = np.zeros_like(self.p)
        self.qux = np.zeros_like(self.p)
        self.qvx = np.zeros_like(self.p)
        self.qex = np.zeros_like(self.p)
        self.qz = np.zeros_like(self.p)
        self.quz = np.zeros_like(self.p)
        self.qvz = np.zeros_like(self.p)
        self.qez = np.zeros_like(self.p)

        # Init K
        self.Kx = np.zeros_like(self.p)
        self.Kux = np.zeros_like(self.p)
        self.Kvx = np.zeros_like(self.p)
        self.Kex = np.zeros_like(self.p)
        self.Kz = np.zeros_like(self.p)
        self.Kuz = np.zeros_like(self.p)
        self.Kvz = np.zeros_like(self.p)
        self.Kez = np.zeros_like(self.p)

        # Initial E & F
        self.Ei = np.zeros_like(self.p)
        self.Eui = np.zeros_like(self.p)
        self.Evi = np.zeros_like(self.p)
        self.Eei = np.zeros_like(self.p)
        self.Fi = np.zeros_like(self.p)
        self.Fui = np.zeros_like(self.p)
        self.Fvi = np.zeros_like(self.p)
        self.Fei = np.zeros_like(self.p)

        if self._cfg.mesh in ['regular', 'adaptative']:
            self.fdtools.Eu(self.Ei, self.Eui, self.Evi, self.Eei,
                            self.r, self.ru, self.rv, self.re, self.p)

        elif self._cfg.mesh == 'curvilinear':
            self.fdtools.EuJ(self.Ei, self.Eui, self.Evi, self.Eei,
                             self.Fi, self.Fui, self.Fvi, self.Fei,
                             self.r, self.ru, self.rv, self.re, self.p,
                             self._msh.dxn_dxp, self._msh.dxn_dzp)

        if self._cfg.mesh in ['regular', 'adaptative']:
            self.fdtools.Fu(self.Fi, self.Fui, self.Fvi, self.Fei,
                            self.r, self.ru, self.rv, self.re, self.p)

        elif self._cfg.mesh == 'curvilinear':
            self.fdtools.FuJ(self.Ei, self.Eui, self.Evi, self.Eei,
                             self.Fi, self.Fui, self.Fvi, self.Fei,
                             self.r, self.ru, self.rv, self.re, self.p,
                             self._msh.dzn_dxp, self._msh.dzn_dzp)

    def init_save(self):
        """ Init save. """

        self.sfile = h5py.File(self._cfg.datafile, 'w')
        self.sfile.create_dataset('x', data=self._msh.x, compression=self._cfg.comp)
        self.sfile.create_dataset('z', data=self._msh.z, compression=self._cfg.comp)
        self.sfile.create_dataset('dx', data=self._msh.dx)
        self.sfile.create_dataset('dz', data=self._msh.dz)
        self.sfile.create_dataset('dt', data=self._cfg.dt)
        self.sfile.create_dataset('nx', data=self._cfg.nx)
        self.sfile.create_dataset('nz', data=self._cfg.nz)
        self.sfile.create_dataset('nt', data=self._cfg.nt)
        self.sfile.create_dataset('ns', data=self._cfg.ns)
        self.sfile.create_dataset('p0', data=self._cfg.p0)
        self.sfile.create_dataset('rho0', data=self._cfg.rho0)
        self.sfile.create_dataset('gamma', data=self._cfg.gamma)
        self.sfile.create_dataset('obstacles', data=self._msh.get_obstacles())
        self.sfile.create_dataset('mesh', data=self._cfg.mesh)
        if self._cfg.probes and self._cfg.probes_loc:
            probes = np.zeros((len(self._cfg.probes_loc), self._cfg.nt))
            self.sfile.create_dataset('probes', data=probes, compression=self._cfg.comp)
            self.sfile.create_dataset('probes_location', data=self._cfg.probes_loc)
        if self._cfg.mesh == 'curvilinear':
            self.sfile.create_dataset('J', data=self._msh.J, compression=self._cfg.comp)
            self.sfile.create_dataset('xn', data=self._msh.xn, compression=self._cfg.comp)
            self.sfile.create_dataset('zn', data=self._msh.zn, compression=self._cfg.comp)
            self.sfile.create_dataset('xp', data=self._msh.xp, compression=self._cfg.comp)
            self.sfile.create_dataset('zp', data=self._msh.zp, compression=self._cfg.comp)

    def get(self):
        """ Get initial fields as a tuple. """
        return self.p, self.r, self.ru, self.rv, self.re
